package complicit

import scala.annotation.tailrec
import scala.collection.JavaConverters._
import scala.collection.mutable.{Buffer, Queue, Map}
import scala.collection.Set
import scala.collection.mutable.HashSet
import scala.util.Random;
import util.control.Breaks._

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import soot.toolkits.scalar.InitAnalysis

import Lattice.Level.{Value => _, _}

class CCTransformer extends BodyTransformer {
  var cfg: UnitGraph = null
  var lg: soot.javaToJimple.LocalGenerator = null
  val gt = GlobalTransform.v
  var body: Body = null

  protected override def internalTransform(body: Body, phase: String,
                                           options: java.util.Map[String, String])
      : Unit = {
    this.body = body
    val clasName = body.getMethod.getDeclaringClass.toString
    if (Main.options.ignoreClasses.contains(clasName)) {
      Util.indented { Util.debug(s"Ignoring class $clasName") }
      return
    }
    if (!GlobalTransform.v.isApplicationMethod(body.getMethod)) {
      Util.indented {
        // Util.debug(s"Skipping non-application method ${body.getMethod}\n")
      }
      return
    }
    if (TaintTransformer.v.isOutputMethod(body.getMethod) != None ||
          TaintTransformer.v.isInputMethod(body.getMethod) != None ||
          gt.isHelperMethod(body.getMethod)) {
      return
    }
    if (clasName.contains("TaintDroidSinkUtil"))
      return
        Util.debug(s"Analyzing method ${body.getMethod} " ++
                     s"[${body.getMethod.getDeclaringClass()}]")
    lg = new soot.javaToJimple.LocalGenerator(body)
    Util.indented {
      handleBody(body)
    }
  }

  def recomputeCFG() = {
    cfg = new BriefUnitGraph(body)
  }

  def handleBody(body : Body): Unit = {
    val units: Queue[SootUnit] = Queue()
    var seen: Set[SootUnit] = Set()
    recomputeCFG()
    Util.debug("Iterating over units:")
    units.enqueue(cfg.getHeads.asScala: _*)
    val allUnits: Set[SootUnit] = body.getUnits.asScala.toSet
    while (!units.isEmpty) {
      breakable {
        val curUnit = units.dequeue()
        if (seen contains curUnit) {
          Util.debug(s"Already seen: $curUnit")
          break
        }
        Util.debug(s"Analyzing unit: $curUnit [${curUnit.getClass}]")
        seen += curUnit
        Util.incIndent()
        handleUnit(curUnit)
        Util.decIndent()
        cfg.getSuccsOf(curUnit).asScala foreach { u =>
          if ((! (seen contains u)) && (allUnits contains u))
            units.enqueue(u)
        }
      }
    }
  }

  def handleUnit(u: SootUnit): Unit = {
    u match {
      case br: IfStmt => handleIf(br)
      case _ => Util.warn(s"CC: Unhandled statement type: $u (${u.getClass})")
    }
  }

  def findAllModified(i: SootUnit, stop: Option[SootUnit]): Set[Value] = {
    var modSeen: Set[SootUnit] = Set()
    def go(i: SootUnit): Set[Value] = {
      if (Some(i) == stop) {
        return Set()
      }
      val init: Set[Value] = i match {
        case as: AssignStmt => Set(as.getLeftOp)
        case _ => Set()
      }
      modSeen += i
      val succs: Seq[SootUnit] = cfg.getSuccsOf(i).asScala filter (!modSeen.contains(_))
      init.union(succs.toSet flatMap go)
    }
    go(i)
  }

  def assignsTo(v: Value, u: SootUnit): Boolean = {
    u match {
      case a: AssignStmt if a.getLeftOp.equals(v) => true
      case _ => false
    }
  }

  /*
  def allPreds(u: SootUnit): Set[SootUnit] = {
    val mhg = new MHGPostDominatorsFinder(cfg)
    def go(u: SootUnit, res: Set[SootUnit]): Set[SootUnit] = {
      if (res.contains(u)) {
        res
      } else {
        val immPreds = cfg.getPredsOf(u).asScala.toSet filter (p => mhg.isDominatedBy(x$1: Unit, x$2: Unit)
        immPreds.union(immPreds flatMap (go(_, res ++ immPreds)))
      }
    }
    go(u, Set())
   }*/

  def isInitialized(v: Value, u: SootUnit): Boolean = {
    val ia: InitAnalysis = new InitAnalysis(cfg)
    v match {
      case l: Local =>
        ia.getFlowBefore(u).contains(l)
      case _ => true
    }
  }

  def createIdAssignment(v: Value, br: SootUnit): Seq[SootUnit] = {
    val mhg = new MHGPostDominatorsFinder(cfg)
    if (!isInitialized(v, br)) {
      // don't generate identity assignments for uninitialized variables
      // these cannot be used outside that branch without the program
      // being rejected by the VM anyway
      return Seq()
    }
    if (gt.upgradeMethods.get(v.getType) == None) {
      return Seq()
    }
    def isFieldRef(v: Value): Boolean = v match {
      case i: InstanceFieldRef => true
      case s: StaticFieldRef => true
      case l: Local => true
      case _ => false
      }
    v match {
      case s if isFieldRef(s) =>
        val tmpLoc = lg.generateLocal(v.getType)
        Seq(Jimple.v.newAssignStmt(tmpLoc, v),
            Jimple.v.newAssignStmt(v, tmpLoc))
      case _ =>
        Util.warn(s"createIdAssignment: Unhandled assignment target: $v")
        Seq()
    }
  }

  def handleIf(br: IfStmt): Unit = {
    val mhg = new MHGPostDominatorsFinder(cfg)
    val dom: Option[SootUnit] = mhg.getImmediateDominator(br) match {
      case null => None
      case x => Some(x)
    }
    val succs = cfg.getSuccsOf(br).asScala map (u => (u, findAllModified(u, dom)))
    Util.debug(s"CC: found modified variables $succs")
    succs foreach { case (s, sMod) =>
      val modifiedInOtherBranches = (succs filter (_._1 != s) flatMap (_._2)) -- sMod
      val assigns = modifiedInOtherBranches flatMap (createIdAssignment(_, br))
      if (!assigns.isEmpty) {
        Util.debug(s"CC: Inserting $assigns")
        body.getUnits.insertBefore(assigns.asJava, s)
        // val nop = Jimple.v.newNopStmt
        // body.getUnits.insertBefore(nop, s)
        s.redirectJumpsToThisTo(assigns.head)
        recomputeCFG()
      }
    }
  }
}

object CCTransformer  {
  private val instance: CCTransformer = new CCTransformer()
  def v: CCTransformer = return instance
}
