package complicit

import soot.jimple._
import soot.{Unit => SootUnit, _}
import scala.io.Source
import sys.process._
import scala.collection.JavaConverters._
import java.io.{File, PrintWriter}


// case class Config(apkMode: Boolean = true, target: String = "ArrayToString1.apk", logOnly: Boolean = false, defaultLow: Boolean = true)
case class Config(apkMode: Boolean = true,
  // target: String = "tainted.apk",
  target: String = "pedometer.apk",
  ignoreClasses: Seq[String] = Seq(),
  logOnly: Boolean = false,
  defaultLow: Boolean = true,
  createOverview: Boolean = true,
  crossCopy: Boolean = false,
  sdkPath: String = s"${Main.homeDir}/taintdroid/android-sdk_eng.daniel_linux-x86")

object Main extends App {

  val toolName = "Complicit"
  val homeDir = System.getProperty("user.home")
  val parser = new scopt.OptionParser[Config](toolName) {
    head(toolName, "0.1")
    arg[String]("<target>") optional() action { (x, c) => c.copy(target = x) } text(
      "Target file for transformation (either a .class or .apk file)")
    opt[Unit]("apk") abbr("a") action ((_, c) => c.copy(apkMode = true)) text(
      "Transform the given APK file")
    opt[Unit]("log") abbr("l") action { (_, c) => c.copy(logOnly = true) } text(
      "Do not apply the actual transformation. Instead, insert calls to Android.log.d"+
      "On each sink. (for debugging purposes).")
    opt[Unit]("quick") abbr("q") action { (_, c) => c.copy(createOverview = false) } text(
      "Do not create an overview")
    opt[Unit]("cross-copy") abbr("c") action { (_, c) => c.copy(crossCopy = true) } text(
      "Add cross-copying statements")
    opt[Seq[String]]("ignore") abbr("i") action { (x, c) => c.copy(ignoreClasses = x) } text(
      "Ignore specified classes (takes a comma-separated list as an argument")
    opt[String]("sdk-path") abbr("s") action { (x, c) => c.copy(sdkPath = x) } text(
      "Path to Android SDK")
    help("help") abbr("h") text("Print this message")
  }
  parser.parse(args, Config()) match {
    case None =>
    case Some(cfg) => run(cfg)
  }

  var options: Config = null

  def run(cfg: Config): Unit = {
    options = cfg
    var target = cfg.target
    // println(cfg)
    if (!options.logOnly) {
      if (options.crossCopy) {
        PackManager.v.getPack("jtp").add(
          new Transform("jtp.ccTransform", CCTransformer.v))
      }
      PackManager.v.getPack("wjtp").add(
        new Transform("wjtp.globalTransform", GlobalTransform.v))
      PackManager.v.getPack("jtp").add(
        new Transform("jtp.taintTransform", TaintTransformer.v))
    } else {
      PackManager.v.getPack("jtp").add(
        new Transform("jtp.logTransform", new LogTransformer()))
    }
    val commonArgs = Array("-w", "-p", "cg", "enabled:falsed")
    Scene.v().addBasicClass("java.lang.System", SootClass.SIGNATURES)
    Scene.v().addBasicClass("java.util.Arrays", SootClass.SIGNATURES)
    if (cfg.apkMode) {
      Scene.v().addBasicClass("android.hardware.Sensor", SootClass.SIGNATURES)
      if (!cfg.target.endsWith(".apk")) {
        Util.die("Target file doesn't end with APK")
      }
      Seq("rm", "-f", s"sootOutput/${cfg.target}").! match {
        case 0 =>
        case _ => { Util.die("Failed to remove old output file sootOutput/${cfg.target}") }
      }
      sys.process.Process(
        Seq("./addClassesToAPK.sh", cfg.target)).! match {
        case 0 => { Util.debug(s"Added helper classes to ${cfg.target}") }
        case _ => { Util.die(s"Failed to add helper classes to ${cfg.target}") }
      }

      var myArgs = Array("-src-prec", "apk", "-f", "dex",
        "-android-jars",
        s"${options.sdkPath}/platforms/", "-allow-phantom-refs")
      if (options.logOnly) {
        // need to delete the old file because soot doesn't allow overwriting
        // with -d
        val oldApk2 = new File("logTransform/" ++ cfg.target)
        if (oldApk2.exists)
          oldApk2.delete
        myArgs = myArgs ++ Array("-d", "logTransform")
        // copy file to not contaminate original apk:
        val replaceExpr = """\.apk$""".r
        val newTargetFile = replaceExpr replaceAllIn (target, _ match {
          case replaceExpr() => "Log.apk"
          case _ => Util.die("Could not find .apk extension")
        })
        Util.debug(s"new target file: $newTargetFile")
        // (too bad java doesn't provide a sane way to copy files without
        // boilerplate...)
        Seq("cp", target, newTargetFile).! match {
          case 0 => Util.debug(s"Copied apk to $newTargetFile")
          case _ => Util.die("cp failed")
        }
        target = newTargetFile
        sys.process.Process(
          Seq("./addLogClass.sh", target),
          new File(".")).! match {
          case 0 => { Util.debug(s"Added helper classes to ${cfg.target}") }
          case _ => { Util.die(s"Failed to add helper classes to ${cfg.target}") }
        }
        val oldApk = new File("logTransform/" ++ target)
        if (oldApk.exists)
          oldApk.delete
      }
      Scene.v().addBasicClass("android.util.Log", SootClass.SIGNATURES);
      soot.Main.main(commonArgs ++ myArgs ++ Array("-process-dir", target))
      if (!options.logOnly) {
        /*
        sys.process.Process(
          Seq("../addClassesToAPK.sh", cfg.target),
          new File("sootOutput")).! match {
          case 0 => { Util.debug(s"Added helper classes to ${cfg.target}") }
          case _ => { Util.die(s"Failed to add helper classes to ${cfg.target}") }
        }
         */
      } else {
        // copy file in logTransform to expected file name
        Seq("cp", "logTransform/" + target, "logTransform/" + cfg.target).! match {
          case 0 => Util.debug(s"Copied output file to ${cfg.target}")
          case _ => Util.die(s"cp failed")
        }
        target = cfg.target
      }
    } else {
      val myArgs = commonArgs ++ Array("-cp", ".", "-pp", "-f", "c",
        "-src-prec", "java")
      soot.Main.main(myArgs ++ Seq(cfg.target))
    }
    if (cfg.createOverview)
      Util.overview.createOverview()
    if (!Util.sound) {
      println("Unsound fallback used, the result may not be secure")
      // System.exit(1)
    }
    if (cfg.apkMode) {
      // We need to resign the APK now
      if (cfg.logOnly) {
        resign(s"logTransform/${target}")
      } else {
        /* Seq("./addPermission.sh", s"sootOutput/${target}",
         "android.permission.SEND_SMS").! match {
          case 0 => Util.debug("Added SMS permission")
          case _ => Util.die("Failed to add SMS permission")
        }
         */
        Seq("./addPermission.sh", s"sootOutput/${target}",
          "android.permission.INTERNET").! match {
          case 0 => Util.debug("Added internet permission")
          case _ => Util.die("Failed to add internet permission")
        }
        resign(s"sootOutput/${target}")
      }
    } else {
      // create script to run file
      val cwd = new File(".").getCanonicalPath
      val script = new File("run.sh")
      script.setExecutable(true)
      // Run file
      // script.getAbsolutePath.!
    }
  }

  /* Resign an apk file */
  def resign(target: String) = {
    Util.debug(s"Attempting to resign $target")
    Seq("zip", "-d", target, "META-INF/*").! match {
      case 0 =>
      case 12 => // this just means there was nothing to remove
      case e => Util.die(s"Removing signatures from ${target} failed with exit code $e")
    }
    Util.debug(s"Removed signatures from $target")
    val homeDir = System.getProperty("user.home")
    Seq("jarsigner", "-verbose", "-sigalg", "SHA1withRSA",
      "-digestalg", "SHA1",
      "-keystore", s"$homeDir/.android/debug.keystore",
      "-storepass", "android", target, "androiddebugkey").! match {
      case 0 => Util.debug(s"Signed ${target}")
      case e => Util.die(s"Signing ${target} failed with exit code $e")
    }
  }
}
